
/***********************************************************************/
/*                             klmod.c                                 */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*    Demoprogramm fuer die Daten des Kaltluftabflussmodells, das von  */
/*  Andreas Schwab programmiert wurde. Als Datenfiles werden im Unter- */
/*  verzeichnis smalldata die Dateien 'klmod_r.gr' (reduzierte Daten)  */
/*  und 'klmod_f.gr' (voller Datensatz) benoetigt. In diesen muss die  */
/*  Fe2d-Instanz mit den Daten fuer den summierten Volumenfluss nach   */
/*  8 Stunden (sumfl.h8) stehen. Die dimension_of_value muss 3 sein,   */
/*  da die fe->z-Koordinaten nicht Null sind.                          */
/*    Die neuen Displaymethoden 'alt-isoline-disp' bzw. 'alt-iso-      */
/*  select-disp' stellen als Funktion die z-Koordinate dar, es ergibt  */
/*  sich also ein farbiges Hoehenprofil bzw. Hoehenlinien.             */
/*    Mit einem Button kann man zwischen dem schnelleren reduzierten   */
/*  Datensatz und dem exakteren vollen Satz umschalten.                */
/*    Ein weiterer Button haengt eine Szene ein, die die Beschriftun-  */
/*  gen der wichtigsten Staedte im Rechengebiet beinhaltet (dazu die   */
/*  Szene am besten nicht rotieren).                                   */
/*    Um die Skalierung der Z-Koordinaten veraendern zu koennen, gibt  */
/*  es zusaetzlich noch einen Ruler.                                   */
/*                                        RALPH SCHWOERER (22.09.95)   */
/***********************************************************************/

#include <grape.h>

#define MAX_POINTS 29008
#define MAX_ELEMENTS 57330
#define MAX_DATA 348096

static  FE2D          *fem, *fe_r, *fe_f;
static  BUTTON        *but_sw_inst;
static  COLORBAR      *colorbar;
static  GRAPHICDEVICE *gdev;
static  int           first = 1, reduced = 1;
static  double        z_scale = 20000.;

extern fe2d_isoline_disp(); /* Nur fuer den Colorbar */





void function_altitude(FE2D *fe, int i, float *bary, float *v)
{ 
  int j;
  
  *v = 0.;
  for(j=0; j<3; j++)
    *v += bary[j] * fe->z[fe->vertex[i][j]] * z_scale;
}





SCENE *scene_switch_instance()
{  
  SCENE      *sc;
  TEXTDEVICE *tdev;
  int        i;

  sc = (SCENE *)START_METHOD(G_INSTANCE);
  ASSURE(sc, "switch-instance: no scene!", END_METHOD(NIL));

  if (first) {
    tdev = (TEXTDEVICE *)GRAPE(InputFile, "new-instance")("smalldata/klmod_f.gr");
    ASSURE(tdev, "can't open file smalldata/klmod_f.gr",exit(1));
    GRAPE(tdev, "set-stddev")();

    fe_f = (FE2D *)GRAPE(Fe2d, "read")();
    ASSURE(fe_r, "can't read Fe2d instance", exit(1));

    first = 0;
  }
  
  if (reduced) {
    fem->number_of_points = fe_f->number_of_points;
    fem->max_number_of_points = fe_f->max_number_of_points;
    fem->number_of_elements = fe_f->number_of_elements;
    fem->max_number_of_elements = fe_r->max_number_of_elements;
    ASSURE((fem->max_number_of_points <= MAX_POINTS &&
	    fem->max_number_of_elements <= MAX_ELEMENTS &&
	    fem->size_of_data <= MAX_DATA),
           "Not enough memory allocated (decrease defines values!)",
	   exit(1));
    memcpy(fem->x,fe_f->x,sizeof(float) * fem->number_of_points);
    memcpy(fem->y,fe_f->y,sizeof(float) * fem->number_of_points);
    for (i=0; i<fem->number_of_points; i++)
      fem->z[i] = fe_f->z[i] / z_scale;
    memcpy(fem->vertex,fe_f->vertex,sizeof(INT3) * fem->number_of_elements);
    memcpy(fem->neighbour,fe_f->neighbour,sizeof(INT3) * fem->number_of_elements);
    fem->f = fe_f->f;
    fem->dimension_of_value = fe_f->dimension_of_value;
    fem->polynomial_order = fe_f->polynomial_order;
    fem->size_of_data = fe_f->size_of_data;
    memcpy(fem->data,fe_f->data,fem->size_of_data);
    strcpy(fem->name,"full");
    GRAPE(but_sw_inst, "set-label")("display -> reduced geometry");          
    reduced = 0;
  }
  else {
    fem->number_of_points = fe_r->number_of_points;
    fem->max_number_of_points = fe_r->max_number_of_points;
    fem->number_of_elements = fe_r->number_of_elements;
    fem->max_number_of_elements = fe_r->max_number_of_elements;
    memcpy(fem->x,fe_r->x,sizeof(float) * fem->number_of_points);
    memcpy(fem->y,fe_r->y,sizeof(float) * fem->number_of_points);
    for (i=0; i<fem->number_of_points; i++)
      fem->z[i] = fe_r->z[i] / z_scale;
    memcpy(fem->vertex,fe_r->vertex,sizeof(INT3) * fem->number_of_elements);
    memcpy(fem->neighbour,fe_r->neighbour,sizeof(INT3) * fem->number_of_elements);
    fem->f = fe_r->f;
    fem->dimension_of_value = fe_r->dimension_of_value;
    fem->polynomial_order = fe_r->polynomial_order;
    fem->size_of_data = fe_r->size_of_data;
    memcpy(fem->data,fe_r->data,fem->size_of_data);
    strcpy(fem->name,"reduced");
    GRAPE(but_sw_inst, "set-label")("display -> full geometry");          
    reduced = 1;
  }

  END_METHOD(sc)
}





FE2D *fe2d_alt_isoline_disp()
{
  FE2D    *fe;
  MANAGER *mgr;
  static  int   logic=0;
  static  float color = 0.7;
  static  float epsilon = -2.0 ;
  static  float offset = 0.0;
  static  RULER *rul[3];
  static  int   add_rul_ref = 0;
  int     i, old_dimval;

  fe = (FE2D *)START_METHOD(G_INSTANCE);
  ASSURE(fe, "", END_METHOD(NIL));

  mgr = (MANAGER *)GRAPE(Manager, "get-stdmgr")();

  colorbar = (COLORBAR *)GRAPE(Colorbar, "get-stdcolorbar")
             (fe2d_isoline_disp, "iso2dclb") ;

  if (logic == 0) {
    rul[0] = (RULER *)GRAPE(Ruler, "new-instance")
             (&color, "range of colormap",RULER_FLOAT);
    rul[1] = (RULER *)GRAPE(Ruler, "new-instance")
             (&epsilon, "ln(line-distance)",RULER_FLOAT);
    rul[2] = (RULER *)GRAPE(Ruler, "new-instance")
             (&offset, "position of colormap",RULER_FLOAT);
    logic = 1;
    GRAPE(colorbar, "button")();
  }

  colorbar->eps = exp(epsilon);
  colorbar->offset = offset;
  colorbar->color = color;
  if(colorbar->display_on_off)     
    GRAPE(colorbar,"display")();

  if (GRAPE(mgr, "new-handle")(fe2d_alt_isoline_disp, 6)) {
    for (i=0; i<3; i++)
      GRAPE(mgr, "add-inter")(rul[i]);
    if (colorbar->button)
      GRAPE(mgr, "add-inter")(colorbar->button);
    if (colorbar->button_options)
      GRAPE(mgr, "add-inter")(colorbar->button_options);
    if (colorbar->button_values)
      GRAPE(mgr, "add-inter")(colorbar->button_values);
  }
  
  old_dimval = fe->dimension_of_value;
  fe->dimension_of_value = 1;

  GRAPE(fe,"isoline")(0, function_altitude, exp(epsilon), color, 
                      offset, 0., colorbar->max,colorbar->min);

  fe->dimension_of_value = old_dimval;

  END_METHOD(fe)
}





FE2D *fe2d_alt_iso_select_disp()
{
  FE2D    *fe;
  MANAGER *mgr;
  static  int   logic=0;
  static  float color = 0.7;
  static  float epsilon = -2.0 ;
  static  float offset = 0.0;
  static  RULER *rul[3];
  static  int   add_rul_ref = 0;
  int     i, old_dimval;

  fe = (FE2D *)START_METHOD(G_INSTANCE);
  ASSURE(fe, "", END_METHOD(NIL));

  mgr = (MANAGER *)GRAPE(Manager,"get-stdmgr")();

  colorbar = (COLORBAR *)GRAPE(Colorbar,"get-stdcolorbar")
             (fe2d_isoline_disp,"iso2dclb") ;

  if (logic == 0) {
    rul[0] = (RULER *)GRAPE(Ruler, "new-instance")
             (&color, "range of colormap",RULER_FLOAT);
    rul[1] = (RULER *)GRAPE(Ruler, "new-instance")
             (&epsilon, "ln(line-distance)",RULER_FLOAT);
    rul[2] = (RULER *)GRAPE(Ruler, "new-instance")
             (&offset, "position of colormap",RULER_FLOAT);
    logic = 1;
    GRAPE(colorbar,"button")();
  }

  colorbar->eps = exp(epsilon);
  colorbar->offset = offset;
  colorbar->color = color;
  if(colorbar->display_on_off)     
    GRAPE(colorbar,"display")();

  if (colorbar->button_values) {
    GRAPE(colorbar->button_values,"set")(ON);
    GRAPE(colorbar->button_values,"set-method")("change-values");
  }

  if (GRAPE(mgr, "new-handle")(fe2d_alt_iso_select_disp, 6)) {
    for (i=0; i<3; i++)
      GRAPE(mgr, "add-inter")(rul[i]);
    if (colorbar->button)
      GRAPE(mgr, "add-inter")(colorbar->button);
    if (colorbar->button_options)
      GRAPE(mgr, "add-inter")(colorbar->button_options);
    if (colorbar->button_values)
      GRAPE(mgr, "add-inter")(colorbar->button_values);
  }
  
  old_dimval = fe->dimension_of_value;
  fe->dimension_of_value = 1;

  GRAPE(fe, "isoline")(1, function_altitude, exp(epsilon), color, 
                       offset, 0., colorbar->max,colorbar->min);

  fe->dimension_of_value = old_dimval;

  END_METHOD(fe)
}





void draw_text(GRAPHICDEVICE *dev, double x, double y, double z, 
               double delta, char *string)
{
  VEC3 koord;
  VEC3 txtcolor = {1.,0.,0.};

  dev->attribute(G_MODE_SET, G_TEXT_COLOR, txtcolor);
  dev->begin_patch();
  dev->patch_color(txtcolor);
  koord[0] = x - delta;
  koord[1] = y - delta;
  koord[2] = z;
  dev->patch_vertex(koord);
  koord[0] = x + delta;
  dev->patch_vertex(koord);
  koord[1] = y + delta;
  dev->patch_vertex(koord);
  koord[0] = x - delta;
  dev->patch_vertex(koord);
  dev->end_patch();

  koord[0] = x + delta + 0.01;
  koord[1] = y;
  koord[2] = z;
  dev->text(koord, string);  
}





SCENE *scene_lettering_disp()
{
  draw_text(gdev, 0.34, 0.49, 0.1,  0.02,  "Freiburg");
  draw_text(gdev, 0.35, 0.68, 0.12, 0.01,  "Emmendingen");
  draw_text(gdev, 1.12, 0.75, 0.1,  0.01,  "Rottweil");
  draw_text(gdev, 1.05, 0.60, 0.1,  0.01,  "Villingen-Schwenningen");
  draw_text(gdev, 1.17, 0.05, 0.08, 0.015, "Schaffhausen");
}





SCENE *scene_add_text()
{  
  SCENE         *sc, *scl, *sco;

  sc = (SCENE *)START_METHOD(G_INSTANCE);
  ASSURE(sc, "add-text: no scene!", END_METHOD(NIL));

  if (strcmp(sc->name,"Lettering") == 0) {
    printf("This scene is already 'Lettering'!\n");
    END_METHOD(sc);
  }
  if (sc->next_scene != NIL) {
    if (strcmp(sc->next_scene->name,"Lettering") == 0) {
      printf("Next Scene is already 'Lettering'!\n");
      END_METHOD(sc);
    }
  }

  scl = (SCENE *)GRAPE(Scene, "new-instance")("Lettering");
  sco = (SCENE *)GRAPE(Scene, "new-instance")("Lettering");
  scl->object = (INSTANCE *)sco;
  scl->next_scene = sc->next_scene;
  scl->method_name = (char *)mem_alloc(sizeof(char) * 40);
  strcpy(scl->method_name, "lettering-disp");
  sc->next_scene = scl;

  END_METHOD(sc)
}





RULER *ruler_scale_z_coord()
{
  int i;

  if (reduced)
    for (i=0; i<fem->number_of_points; i++)
      fem->z[i] = fe_r->z[i] / z_scale;
  else
    for (i=0; i<fem->number_of_points; i++)
      fem->z[i] = fe_f->z[i] / z_scale;

  END_METHOD(NIL)
}




main()
{  
  SCENE      *sc;
  MANAGER    *mgr;
  TEXTDEVICE *tdev;
  BUTTON     *button_text;
  RULER      *rul_scale_z;
  VEC3       xyz;
  int        i;

  printf("\n\n*----------------------------------------------------*");
  printf(  "\n|                                                    |");
  printf(  "\n|   +  Choose one of the 'Scene display method's:    |");
  printf(  "\n|                                                    |");
  printf(  "\n|      - alt-isoline                                 |");
  printf(  "\n|      - alt-iso-selct                               |");
  printf(  "\n|      - vect (distance: -0.001, mesh with: -3.5)    |");
  printf(  "\n|                                                    |");
  printf(  "\n|   +  With the buttons in the 'opts' you can        |");
  printf(  "\n|                                                    |");
  printf(  "\n|      - get more exact datas (but slower)           |");
  printf(  "\n|      - get letterings of important towns           |");
  printf(  "\n|                                                    |");
  printf(  "\n*----------------------------------------------------*\n\n");

  GRAPE(Scene, "add-method")("switch-instance", scene_switch_instance);
  GRAPE(Fe2d ,"add-method")("alt-isoline-disp", fe2d_alt_isoline_disp);
  GRAPE(Fe2d, "add-method")("alt-iso-select-disp", fe2d_alt_iso_select_disp);
  GRAPE(Scene, "add-method")("add-text", scene_add_text);
  GRAPE(Scene, "add-method")("lettering-disp", scene_lettering_disp);
  GRAPE(Ruler, "add-method")("scale-z-coord", ruler_scale_z_coord);

  mgr = (MANAGER *)GRAPE(Manager, "get-stdmgr")();
  gdev = (GRAPHICDEVICE *)GRAPE(GraphicDevice, "get-stddev")();
   
  tdev = (TEXTDEVICE *)GRAPE(InputFile, "new-instance")("klmod0.st");
  ASSURE(tdev, "can't open file klmod0.st", exit(1));
  GRAPE(tdev, "set-stddev")();

  GRAPE(gdev, "read")();
  g_ctl_set_surfaces();
  g_ctl_set_lights ();
  g_ctl_set_trans ();
  GRAPE(mgr, "set-graphic-redraw")(ON);

  tdev = (TEXTDEVICE *)GRAPE(InputFile, "new-instance")("smalldata/klmod_r.gr");
  ASSURE(tdev, "can't open file smalldata/klmod_r.gr", exit(1));
  GRAPE(tdev, "set-stddev")();

  fe_r = (FE2D *)GRAPE(Fe2d, "read")();
  ASSURE(fe_r,"can't read Fe2d instance",exit(1));

  fem = (FE2D *)GRAPE(Fe2d, "new-instance")("Fe_main");
  fem->number_of_points   = fem->max_number_of_points   = MAX_POINTS;
  fem->number_of_elements = fem->max_number_of_elements = MAX_ELEMENTS;
  fem->x         = (float *)mem_alloc(sizeof(float) * fem->number_of_points);
  fem->y         = (float *)mem_alloc(sizeof(float) * fem->number_of_points);
  fem->z         = (float *)mem_alloc(sizeof(float) * fem->number_of_points);
  fem->vertex    = (INT3 *)mem_alloc(sizeof(INT3) * fem->number_of_elements);
  fem->neighbour = (INT3 *)mem_alloc(sizeof(INT3) * fem->number_of_elements);
  fem->data      = (char *)mem_alloc(MAX_DATA);

  fem->number_of_points = fe_r->number_of_points;
  fem->max_number_of_points = fe_r->max_number_of_points;
  fem->number_of_elements = fe_r->number_of_elements;
  fem->max_number_of_elements = fe_r->max_number_of_elements;
  memcpy(fem->x,fe_r->x,sizeof(float) * fem->number_of_points);
  memcpy(fem->y,fe_r->y,sizeof(float) * fem->number_of_points);
  for (i=0; i<fem->number_of_points; i++)
    fem->z[i] = fe_r->z[i] / z_scale;
  memcpy(fem->vertex,fe_r->vertex,sizeof(INT3) * fem->number_of_elements);
  memcpy(fem->neighbour,fe_r->neighbour,sizeof(INT3) * fem->number_of_elements);
  fem->f = fe_r->f;
  fem->dimension_of_value = fe_r->dimension_of_value;
  fem->polynomial_order = fe_r->polynomial_order;
  fem->size_of_data = fe_r->size_of_data;
  memcpy(fem->data,fe_r->data,fem->size_of_data);
  strcpy(fem->name,"reduced");
  

  sc = (SCENE *)GRAPE(Scene, "new-instance")("Relief");
  sc->object = (INSTANCE *)fem;

  GRAPE(fe_r, "isoline-disp")();   
  /*  Damit die Routine fline_select() der "alt-iso-select-disp"-Methode
      die colorbar->values kennt!  */

  but_sw_inst = (BUTTON *)GRAPE(Button, "new-instance")
                ("switch-instance", sc, "Display -> Full Geometry");
  GRAPE(but_sw_inst, "set-pref-size")(10.5, 1.5);          
  GRAPE(but_sw_inst, "set-pref-pos")(0.5, 1.8);          
  GRAPE(but_sw_inst, "set-color-index")(4);        

  button_text = (BUTTON *)GRAPE(Button, "new-instance")
                ("add-text", NIL, "Add Scene 'Lettering'");
  GRAPE(button_text, "set-pref-size")(10.5, 1.5);          
  GRAPE(button_text, "set-pref-pos")(0.5, 3.8);          

  rul_scale_z = (RULER *)GRAPE(Ruler, "new-instance")
                (&z_scale, "Divisor of Z-Coordinate", RULER_DOUBLE);
  GRAPE(rul_scale_z, "set-method")("scale-z-coord");
  GRAPE(rul_scale_z, "set-pref-size")(10.5, 1.0);
  GRAPE(rul_scale_z, "set-pref-pos")(0.5, 5.3);
  GRAPE(rul_scale_z, "set-scale")(2. * z_scale);
  GRAPE(rul_scale_z, "set-offset")(0.);
  GRAPE(rul_scale_z, "set-step-size")(1000.);

  GRAPE(mgr, "add-inter")
       (new_item(Group, I_Name, "Scene Manipulations", I_Border, bfBorder, 
	         I_SizeY, 6.1, I_Item, but_sw_inst, I_Item, rul_scale_z,
	         I_Item, button_text, I_End));

  xyz[0] = -0.9;
  xyz[1] = -0.9;
  xyz[2] = 0.;
  GRAPE(Colorbar, "set-default")(xyz, 0.4, 0.1, 1, 100.,
	1500., 0, NIL, fe2d_isoline_disp, "iso2dclb"); 

  GRAPE(mgr, "handle")(sc);
}   
   





























